function delta=EM_b_error(E,a)
% Computes backward error on polynomial coefficients 
% according to Edelman-Murakami (Thm 2.1, in matrix form).
%
% Input:
% - perturbation matrix E,
% - column vector a of polynomial coefficients
%   (starting from constant term, leading coeff. must be 1).
%
% Output:
% - column vector delta of perturbations.

% Compute polynomial degree.
n=length(a)-1;
% Form matrix.
A=zeros(n,n+1);
% Fill upper part of A.
for d=0:n-1
    s=0;
    for ii=1:n-d
    s=s+E(ii,ii+d);
    A(ii,ii+d+1)=-s;
    end
end
for d=-1:-1:-(n-1)
    s=0;
    for ii=1-d:n
        s=s+E(ii,ii+d);
        A(ii-1,ii+d)=s;
    end
end
delta=A*a;
    