function [sigmaanew,gvnew,hvnew,bvnew,z1,w1,db1,gunew,hunew,bunew,p1,q1]=structured_QZ_infin(sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q,kin,kfin,Qin);
% Implementation of Theorem 6.1 in the draft.
% Recall that:
% - sigma,gv,hv,bv,z,w are parameters for A,
% - d,gu,hu,bu,p,q are parameters for B,
% - Qin is the initial Givens matrix (encoding the shift).

% Recover matrix size:
N=length(z); % assuming everything is consistent! 

% make sure that q and w are in column form
[ws1,ws2]=size(w);
if length(w)>ws1
    w=w.';
end
[qs1,qs2]=size(q);
if length(q)>qs1
    q=q.';
end

% *** CHECK ***
% A=recover_A(sigmaa,gv,hv,bv,z,w);
% B=recover_B(db,gu,hu,bu,p,q);
% V=[zeros(1,N-1),1;eye(N-1),zeros(N-1,1)];
% U=eye(N);
% *************

% Initialize new parameters:
sigmaanew=zeros(N-1,1);
gvnew=zeros(N,2);
hvnew=zeros(2,N);
bvnew=zeros(2,2,N-1);
z1=zeros(N,1);
w1=zeros(N,1);
dbnew=zeros(N,1);
gunew=zeros(N,1);
hunew=zeros(1,N);
bunew=zeros(1,N-1);
p1=zeros(N,1);
q1=zeros(N,1);

% Initialize auxiliary parameters:
du=zeros(N,1);
du1tilde=zeros(N+1,1);
gu1tilde=zeros(N+1,2);
hu1tilde=zeros(2,N+1);
bu1tilde=zeros(2,2,N+1);
dv1tilde=zeros(N+1,1);
gv1tilde=zeros(N+1,3);
hv1tilde=zeros(3,N+2);
bv1tilde=zeros(3,3,N+1);
du1=zeros(N,1);
gu1=zeros(N,2);
hu1=zeros(2,N);
bu1=zeros(2,2,N);
dv1=zeros(N+1,1);
gv1=zeros(N+1,3);
hv1=zeros(3,N+1);
bv1=zeros(3,3,N+1);
%epsilon=zeros(N,1);

% Implement Theorem 6.1:

% 1)
%sigmav=sigmaa(1)+z(2)*w(1)'; % was sigmav1
sigmav=sigmaa+z(2:N).*conj(w(1:N-1));
epsi=gv(1,:)*hv(:,1)-z(1)*w(1)';
%Q=planerot([epsi-alpha;sigmaa(1)]); % was ^* in the draft
if kin==1
    Q=Qin;
else
    Q=eye(2);
end

Gamma=Q*[gv(1,:)*hv(:,1), gv(1,:)*bv(:,:,1);sigmav(1), gv(2,:)];
gv1tilde(2,:)=Gamma(1,:);   
fv=Gamma(2,1);
phiv=Gamma(2,2:3);
auxvec=Q*z(1:2);
z1(1)=auxvec(1);
chi=auxvec(2);
fa=fv-chi*w(1)';
phia=phiv;

% *** CHECK ***
% A(1:2,:)=Q*A(1:2,:);
% V(1:2,:)=Q*V(1:2,:);
% *************

% 2)
gamma=w(1);
c=p(1);
theta=q(1);
du(1)=db(1)+p(1)*q(1)';
fu=du(1);
phiu=gu(1);
fb=db(1);
phi=gu(1);

% 3)
for k=1:N-2
% 3a)    
vareps=phi*hu(k+1)-c*q(k+1)';
epsi=phia*hv(:,k+1)-chi*w(k+1)';
du(k+1)=db(k+1)+p(k+1)*q(k+1)';
Phi=Q*[fb,vareps;0,db(k+1)];

% *** CHECK ***
% B(k:k+1,:)=Q*B(k:k+1,:);
% U(k:k+1,:)=Q*U(k:k+1,:);
% *************

% 3b)
if k>=kin
    G=planerot(Phi(2,:)');
    Z=fliplr(G');
else
    Z=eye(2);
end
Omega=[fa,epsi;0,sigmaa(k+1)]*Z;

% *** CHECK ***
% A(:,k:k+1)=A(:,k:k+1)*Z;
% V(:,k:k+1)=V(:,k:k+1)*Z;
% B(:,k:k+1)=B(:,k:k+1)*Z;
% U(:,k:k+1)=U(:,k:k+1)*Z;
% *************

% 3c)
if (k>=kin)&&(k<=kfin-2)
  [newQ,yy]=planerot(Omega(:,1));  
  sigmaanew(k)=yy(1);
else
    if k+1==kin
        newQ=Qin;
    else
        newQ=eye(2);
    end
    yy=newQ*Omega(:,1);
    sigmaanew(k)=yy(1);
end
  
% *** CHECK ***
% A(k+1:k+2,:)=newQ*A(k+1:k+2,:);
% V(k+1:k+2,:)=newQ*V(k+1:k+2,:);
% *************

% 3d)
Gammap=newQ*[fv,phiv*hv(:,k+1),phiv*bv(:,:,k+1);z(k+2)*gamma',sigmav(k+1),gv(k+2,:)];
Gamma=Gammap*[Z,zeros(2);zeros(2),eye(2)];
C=[1,0,zeros(1,2);zeros(2,1),hv(:,k+1),bv(:,:,k+1)]*[Z,zeros(2);zeros(2),eye(2)];
dv1tilde(k+2)=Gamma(1,1);
gv1tilde(k+2,:)=Gamma(1,2:4);
betav=Gamma(2,2:4);
fv=betav(1);
phiv=betav(2:3);
hv1tilde(:,k+2)=C(:,1);
bv1tilde(:,:,k+2)=C(:,2:4);
% 3e)
Lambda=Q*[fu,phiu*hu(k+1),phiu*bu(k+1);p(k+1)*theta',du(k+1),gu(k+1)]*[Z,zeros(2,1);zeros(1,2),1];
D=[1,0,0;0,hu(k+1),bu(k+1)]*[Z,zeros(2,1);zeros(1,2),1];
du1tilde(k)=Lambda(1,1);
gu1tilde(k,:)=Lambda(1,2:3);
betau=Lambda(2,2:3);
fu=betau(1);
phiu=betau(2);
hu1tilde(:,k+1)=D(:,1);
bu1tilde(:,:,k+1)=D(:,2:3);
% 3f)
%auxvec=newQ*[chi;z(k+1)];
auxvec=newQ*[chi;z(k+2)];
z1(k+1)=auxvec(1);
chi=auxvec(2);
auxvec=Z'*[gamma;w(k+1)];
w1(k)=auxvec(1);
gamma=auxvec(2);
% 3g)
auxvec=Z'*[theta;q(k+1)];
q1(k)=auxvec(1);
theta=auxvec(2);
auxvec=Q*[c;p(k+1)];
p1(k)=auxvec(1);
c=auxvec(2);
% 3h)
fa=fv-chi*gamma';
phia=phiv;
fb=fu-c*theta';
phi=phiu;
% update Q
Q=newQ;
end

% 4)
vareps=phi*hu(N)-c*q(N)';
du(N)=db(N)+p(N)*q(N)';
Phi=Q*[fb,vareps;0,db(N)];

% *** CHECK ***
% B(N-1:N,:)=Q*B(N-1:N,:);
% U(N-1:N,:)=Q*U(N-1:N,:);
% *************

% 5)
G=planerot(Phi(2,:)');
Z=fliplr(G');

% 6)
Gamma=[fv, phiv*hv(:,N)]*Z;
C=[1,0;zeros(2,1),hv(:,N)]*Z; 
hvtilde1=1;  
dv1tilde(N+1)=Gamma(1);
gv1tilde(N+1,1)=Gamma(2);  % this should be a number, padding w/zero
hv1tilde(:,N+1)=C(:,1);
bv1tilde(:,1,N+1)=C(:,2);

% *** CHECK ***
% A(:,N-1:N)=A(:,N-1:N)*Z;
% V(:,N-1:N)=V(:,N-1:N)*Z;
% B(:,N-1:N)=B(:,N-1:N)*Z;
% U(:,N-1:N)=U(:,N-1:N)*Z;
% *************

% 7)
Lambda=Q*[fu,phiu*hu(N);p(N)*theta',du(N)]*Z;
D=[1,0;0,hu(N)]*Z;
hutilde1=1;
du1tilde(N-1)=Lambda(1,1);
gu1tilde(N-1,1)=Lambda(1,2);  % this should be a number, padding w/zero
du1tilde(N)=Lambda(2,2);
hu1tilde(:,N)=D(:,1);
bu1tilde(:,1,N)=D(:,2);
hv1tilde(1,N+2)=1; % this should be a number, padding w/zero
hu1tilde(1,N+1)=1; % this should be a number, padding w/zero

% 8)
w1(N-1:N)=Z'*[gamma;w(N)];
p1(N-1:N)=Q*[c;p(N)];
q1(N-1:N)=Z'*[theta;q(N)];
z1(N)=chi;
sigmaanew(N-1)=dv1tilde(N+1)-z1(N)*w1(N-1)';

% 9)
db1=du1tilde(1:N)-p1.*conj(q1); 

% 10)
gv1=gv1tilde(2:N+1,:);
hv1=zeros(3,N);
hv1(:,1:N)=hv1tilde(:,3:N+2);  
bv1=zeros(3,3,N);
bv1(:,:,1:N-1)=bv1tilde(:,:,3:N+1);
dv1(2:N)=dv1tilde(3:N+1);

gu1=gu1tilde(1:N,:);   
hu1=hu1tilde(:,2:N+1);  
bu1(:,:,2:N-1)=bu1tilde(:,:,3:N);   
du1=du1tilde(1:N); 

% now we have redundant generators for the updated A and B

% 11)  compression of V
bs=zeros(2,2,N); 
gs=zeros(N,2);
hs=zeros(2,N);

Sold=gv1(1,:);
gs(1,1)=1;
hs(:,1)=Sold*hv1(:,1);  % this settles the (1,1) entry

gs(2,:)=[0,1];
bs(:,:,1)=[1,0; 0,0];
Snew=[Sold*bv1(:,:,1);gv1(2,:)];
hs(:,2)=Snew*hv1(:,2);

z=[gs(1,:)*hs(:,1);dv1(2)];
yold=w1(1);

for k=2:N-1
    [Vv,Y]=planerotr([yold',w1(k)']);
    ynew=Y(2);
    Aa=[z, Snew*hv1(:,k); z1(k+1)*yold', dv1(k+1)]*Vv;
    yold=ynew;
    auxvec=Aa(:,1); % to be reduced by F
    [F,ww]=hh3(auxvec);
    Sold=Snew;
    Snew=[Sold*bv1(:,:,k);gv1(k+1,:)];
    Bb=F*[Aa(:,2),Snew];
    z=Bb(2:3,1);
    
    bs(:,:,k)=F(2:3,1:2)';
    gs(k+1,:)=F(2:3,3)';
    Snew=gs(k+1,:)'*gv1(k+1,:)+bs(:,:,k)'*Sold*bv1(:,:,k);
    hs(:,k+1)=Snew*hv1(:,k+1);
end
gvnew=gs;
hvnew=hs;
bvnew=bs;


% 12)  compression of U
X=q1(1);
Y=gu1(1,:);
zed=du1(1);
hunew(:,2)=gu1(1,:)*hu1(:,2);   
gunew(1)=1; 
for k=2:N-1   
    [WW,auxv]=planerotr([X',q1(k)']);
    Zed=[zed,hunew(:,k);p1(k)*X',du1(k)]*WW;  
    X=auxv(2);
    hp=Zed(1,2);
    hpp=Zed(2,2);
    Effe=(planerot(Zed(:,1)))'; % F
    bunew(k)=Effe(1,2);
    gunew(k)=Effe(2,2);
    Y=gunew(k)'*gu1(k,:)+bunew(k)'*Y*bu1(:,:,k);
    zed=gunew(k)'*hpp+bunew(k)'*hp;
    hunew(k+1)=Y*hu1(:,k+1);
end
%*** CHECK *****
% aa=recover_A(sigmaanew,gvnew,hvnew,bvnew,z1,w1);
% bb=recover_B(db1,gunew,hunew,bunew,p1,q1);
% na=norm(A-aa);
% nb=norm(B-bb);
% display(na-nb)
%***************